package require Tcl 8.6
package require json 1.3
package require struct::list 1.8

package provide bills 1.0

namespace eval bills {

    proc parseSplit {data dictName} {
        upvar $dictName splits

        set name [dict get $data name]
        if {![dict exists $splits $name]} {
            dict set splits $name [list \
                text        $name \
                sum         0 \
                quantity    0 \
                manual      false
            ]
        }
        dict with splits $name {
            set sum [expr {$sum + [dict get $data sum] / 100.0}]
            set quantity [expr {$quantity + [dict get $data quantity]}]
        }
    }

    proc parseBill {data} {
        set dateTime [dict get $data dateTime]
        if {![regexp {^[0-9]+$} $dateTime]} {
            # does not look like UNIX time...
            set dateTime [clock scan $dateTime \
                -format %Y-%m-%dT%H:%M:%S \
                -gmt yes \
            ]
        }
        # Decrement timeval to Europe/Moscow time offset. This offset is added
        # for unknown reason to both unixtime and ISO-like timestamp formats.
        set dateTime [clock add $dateTime -3 hours]

        # parse shop name
        if {[dict exists $data retailPlace]} {
            set shopName [string trim [dict get $data retailPlace]]
        } else {
            set shopName ""
        }
        if {$shopName in {"" "null"} && [dict exists $data user]} {
            set shopName [string trim [dict get $data user]]
        }
        if {$shopName in {"" "null"}} {
            set shopName [string trim [dict get $data userInn]]
        }
        if {$shopName in {"" "null"}} {
            set shopName -
        }

        # parse splits
        set splits {}
        set i 1
        foreach item [dict get $data items] {
            try {
                parseSplit $item splits
            } on error {err opts} {
                dict set opts -errorinfo "split $i: [dict get $opts -errorinfo]"
                dict set opts -errorcode {BILL2QIF LOGIC}
                return -options $opts [mc "split %d: %s" $i $err]
            }
            incr i
        }

        list \
            dateTime    $dateTime \
            sum         [expr {[dict get $data totalSum] / 100.0}] \
            shopName    $shopName \
            TIN         [dict get $data userInn] \
            splits      [dict values $splits]
    }

    proc parseData {json} {
        if {[catch {set data [json::json2dict $json]} err opts]} {
            error $err $::errorInfo {BILL2QIF INPUTFORMAT}
        }
        if {[llength $data] == 0} {
            throw {BILL2QIF LOGIC} [mc "empty bill list"]
        }

        if {[llength $data] % 2 == 0 && [dict exists $data totalSum]} {
            # looks like a single-bill file
            try {
                set res [list [parseBill $data]]
            } on error {err opts} {
                dict set opts -errorcode {BILL2QIF LOGIC}
                return -options $opts $err
            }
            return $res
        } else {
            # parse as multiple-bill file
            set res {}
            set i 1
            foreach doc $data {
                try {
                    lappend res [parseBill [dict get $doc document receipt]]
                } on error {err opts} {
                    dict set opts -errorinfo "bill $i: [dict get $opts -errorinfo]"
                    dict set opts -errorcode {BILL2QIF LOGIC}
                    return -options $opts [mc "bill %d: %s" $i $err]
                }
                incr i
            }
            return $res
        }
    }

    namespace export *

}

# vim: set ft=tcl:

